using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using Mapbox.Unity.Location;
using Mapbox.Unity.Map;
using Mapbox.Utils;
using System.Linq;
using System.IO;
using System;
using UnityEngine.UI;

public class DataPointVisualize : MonoBehaviour
{
  public AbstractMap map;
  public GameObject dataPathPrefab;
  public int dataLimit = 1000;
  public DataPointProvider dataPointProvider;

  ILocationProvider _locationProvider;

  ILocationProvider LocationProvider
  {
    get
    {
      if (_locationProvider == null)
      {
        _locationProvider = LocationProviderFactory.Instance.DefaultLocationProvider;
      }

      return _locationProvider;
    }
  }

  Vector3 _targetPosition;

  public List<DataPath> Paths
  {
    get
    {
      return paths;
    }
  }

  List<DataPath> paths = new List<DataPath>();

  public void SetTime(float t)
  {
    var dataPaths = GameObject.FindObjectsOfType<DataPath>();
    for (int j = 0; j != dataPaths.Length; ++j)
    {
      var lr = dataPaths[j].gameObject.GetComponent<LineRenderer>();
      var material = lr.material;
      material.SetFloat("_TimeF", t);
    }

    long currentEpoch = Lerp(SceneConfiguration.Instance.minTimeStamp, SceneConfiguration.Instance.maxTimeStamp, t);
    SceneConfiguration.Instance.textCurrentTime.text = ToEpoch(currentEpoch).ToShortTimeString();
  }

  float timeRangeLeft = 0;
  float timeRangeRight = 1;

  public void SetTimeRangeLeft(float l)
  {
    timeRangeLeft = l;
    SetTimeRange(timeRangeLeft, timeRangeRight);
  }

  public void SetTimeRangeRight(float r)
  {
    timeRangeRight = r;
    SetTimeRange(timeRangeLeft, timeRangeRight);
  }

  public void SetTimeRange(float l, float r)
  {
    if (r < l) {
      swap(ref l, ref r);
    }

    var dataPaths = GameObject.FindObjectsOfType<DataPath>();
    for (int j = 0; j != dataPaths.Length; ++j)
    {
      var dataPath = dataPaths[j];
      var lr = dataPath.gameObject.GetComponent<LineRenderer>();
      //var material = lr.material;
      //material.SetFloat("_TimeF", l); //TODO: Shader must be optimized to allow for range

      var allPoints = dataPath.positions;
      int oldCount = allPoints.Length;
      int start = (int)(oldCount * l);
      int end = (int)(oldCount * r);
      int count = end - start;

      if (count > lr.positionCount) {
        lr.positionCount = count;
      }

      if (start != 0) {
        for (int i = 0; i != count; ++i) {
          lr.SetPosition(i, allPoints[i + start]);
        }
      }
      lr.positionCount = count;
    }

    SetEpochTimeForTextBox(l, SceneConfiguration.Instance.textCurrentTimeLeft);
    SetEpochTimeForTextBox(r, SceneConfiguration.Instance.textCurrentTimeRight);
  }

  private void swap(ref float l, ref float r)
  {
    var tmp = r;
    r = l;
    l = tmp;
  }

  private void SetEpochTimeForTextBox(float range, Text textBox)
  {
    long currentEpoch = Lerp(SceneConfiguration.Instance.minTimeStamp, SceneConfiguration.Instance.maxTimeStamp, range);
    textBox.text = ToEpoch(currentEpoch).ToShortTimeString();
  }

  private long Lerp(long min, long max, float t)
  {
    return min + (long)((max - min) * t);
  }

  void Start()
  {
    LocationProviderFactory.Instance.mapManager.OnInitialized += () =>  {
      var objs = dataPointProvider.pointList;

      int lastID = 0;
      DataPath path = null;
      int pathCount = 0;

      for (int i = 0; i < objs.Length; i++)
      {
        var point = objs[i];

        SceneConfiguration.Instance.minTimeStamp = Math.Min(SceneConfiguration.Instance.minTimeStamp, point.timestamp);
        SceneConfiguration.Instance.maxTimeStamp = Math.Max(SceneConfiguration.Instance.maxTimeStamp, point.timestamp);

        if (lastID != point.id_hash) {

          if(path != null) {
            path.Setup();
          }

          if(pathCount == this.dataLimit) break;

          lastID = point.id_hash;

          GameObject go = Instantiate(this.dataPathPrefab, this.transform);
          go.name = point.id_hash.ToString();
          path = go.GetComponent<DataPath>();
          path.id_hash = point.id_hash;

          paths.Add(path);

          pathCount++;
        }

        path.points.Add(point);
      }

      path.Setup();

      SceneConfiguration.Instance.textMinTimeStamp.text = ToTimeString(SceneConfiguration.Instance.minTimeStamp);
      SceneConfiguration.Instance.textMaxTimeStamp.text = ToTimeString(SceneConfiguration.Instance.maxTimeStamp);

      Debug.Log("dataPath count = " + GameObject.FindObjectsOfType<DataPath>().Length);

      SetTime(0);
    };
    
  }

  static public DateTime ToEpoch(long epoch)
  {
    return new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc).AddSeconds(epoch);
  }

  private string ToTimeString(long epoch)
  {
    DateTime dateTime = ToEpoch(epoch);

    return dateTime.DayOfWeek.ToString() + " " + dateTime.ToShortTimeString();
  }

  private int isign(long s)
  {
    if(s < 0) return -1;
    if(s > 0) return 1;
    return 0;
  }

  public void OnGUI()
  {

    // Starts a vertical group
    GUILayout.BeginVertical("box");

    if (GUILayout.Button("Y = Time"))
    {
      ConvertToSpaceTimePath();
    }
    GUILayout.Space(20);
    for (int i = 0; i < paths.Count; i++)
    {
      var path = paths[i];
      bool active = path.GetComponent<LineRenderer>().enabled;
      string text = "[" + (active ? "X" : "_") + "] " + path.id_hash;

      if (GUILayout.Button(text)) {
        path.GetComponent<LineRenderer>().enabled = !active;
      }
    }
    GUILayout.EndVertical();

  }

  private void ConvertToSpaceTimePath()
  {
    for (int i = 0; i < paths.Count; i++)
    {
      var path = paths[i];
      var lr = path.GetComponent<LineRenderer>();



      int n = lr.positionCount;
      var points = new Vector3[n];
      lr.GetPositions(points);
      for (int j = 0; j != n; ++j)
      {
        points[j].y = ((float)j / n) * 40;
      }
      lr.SetPositions(points);
    }
  }
}
